import ast

import pytest
from vulture import core


def check(items_or_names, expected_names):
    if isinstance(items_or_names, set):
        # items_or_names is a set of strings.
        assert items_or_names == set(expected_names)
    else:
        # items_or_names is a list of Item objects.
        names = sorted(item.name for item in items_or_names)
        expected_names = sorted(expected_names)
        assert names == expected_names


def check_unreachable(v, lineno, size, name):
    assert len(v.unreachable_code) == 1
    item = v.unreachable_code[0]
    assert item.lineno == lineno
    assert item.size == size
    assert item.name == name


def skip_if_not_has_async(function):
    if not hasattr(ast, 'AsyncFunctionDef'):
        pytest.mark.skip(
            function, reason="needs async support (added in Python 3.5)")


@pytest.fixture
def v():
    return core.Vulture(verbose=True, sort_by_size=True)



import ast
import codecs
import os
import tokenize

# Encoding to use when converting input files to unicode.
ENCODING = 'utf-8'


class VultureInputException(Exception):
    pass


def condition_is_always_true(condition):
    try:
        return _evaluate_condition(condition)
    except ValueError:
        return False


def condition_is_always_false(condition):
    try:
        return not _evaluate_condition(condition)
    except ValueError:
        return False


def _evaluate_condition(condition):
    """
    Try to safely evaluate the given condition. Return True or False if
    the if the given condition is always True or False, respectively.
    Raise ``ValueError`` if the condition cannot be evaluated safely.

    The evaluation will only succeed if the condition exclusively
    consists of Python literals. We could use eval() to catch more
    cases. However, this function is not safe for arbitrary Python code.
    Even after overwriting the "__builtins__" dictionary, the original
    dictionary can be restored
    (https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html).

    """
    try:
        result = bool(ast.literal_eval(condition))
    except ValueError:
        raise ValueError("Condition cannot be evaluated")
    else:
        return result


def format_path(path):
    if not path:
        return path
    relpath = os.path.relpath(path)
    return relpath if not relpath.startswith('..') else path


def read_file(filename):
    # Python >= 3.2
    try:
        # Use encoding detected by tokenize.detect_encoding().
        with tokenize.open(filename) as f:
            return f.read()
    except (SyntaxError, UnicodeDecodeError) as err:
        raise VultureInputException(err)
    except AttributeError:
        # tokenize.open was added in Python 3.2.
        pass

    # Python < 3.2
    try:
        with codecs.open(filename, encoding=ENCODING) as f:
            return f.read()
    except UnicodeDecodeError as err:
        raise VultureInputException(err)


class LoggingList(list):
    def __init__(self, typ, verbose):
        self.typ = typ
        self._verbose = verbose
        return list.__init__(self)

    def append(self, item):
        if self._verbose:
            print('define {0} "{1}"'.format(self.typ, item.name))
        list.append(self, item)


class LoggingSet(set):
    def __init__(self, typ, verbose):
        self.typ = typ
        self._verbose = verbose
        return set.__init__(self)

    def add(self, name):
        if self._verbose:
            print('use {0} "{1}"'.format(self.typ, name))
        set.add(self, name)



from . import v
assert v  # Silence pyflakes.


def test_syntax_error(v):
    v.scan("foo bar")
    assert int(v.report()) == 1


def test_null_byte(v):
    v.scan("\x00")
    assert int(v.report()) == 1



import sys

from vulture import core

dc = core.DEFAULT_CONFIDENCE


def check_min_confidence(code, min_confidence, expected):
    v = core.Vulture(verbose=True, min_confidence=min_confidence)
    v.scan(code)
    detected = dict((item.name, item.confidence)
                    for item in v.get_unused_code())
    assert detected == expected


def test_confidence_import():
    code = """\
import foo
"""
    check_min_confidence(code, 50, {'foo': 90})
    check_min_confidence(code, 100, {})


def test_confidence_unreachable():
    code = """\
def foo():
    return
    bar()

foo()
"""
    check_min_confidence(code, 50, {'return': 100})
    check_min_confidence(code, 100, {'return': 100})


def test_function_arg():
    code = """\
def foo(a):
    b = 3

foo(5)
"""
    if sys.version_info < (3, 0):
        # Python 2
        check_min_confidence(code, 50, {'a': dc, 'b': dc})
        check_min_confidence(code, dc, {'a': dc, 'b': dc})
        check_min_confidence(code, 100, {})
    else:
        # Python 3
        check_min_confidence(code, 50, {'a': 100, 'b': dc})
        check_min_confidence(code, dc, {'a': 100, 'b': dc})
        check_min_confidence(code, 100, {'a': 100})


def test_confidence_class():
    code = """\
class Foo:
    pass
"""
    check_min_confidence(code, 50, {'Foo': dc})
    check_min_confidence(code, 100, {})


def test_confidence_attr():
    code = "A.b = 'something'"
    check_min_confidence(code, 50, {'b': dc})
    check_min_confidence(code, 100, {})


def test_confidence_props():
    code = """\
class Foo:
    @property
    def some_prop():
        pass

Foo()
"""
    check_min_confidence(code, 50, {'some_prop': dc})
    check_min_confidence(code, 100, {})



from . import check, v
assert v  # Silence pyflakes.


def test_function_object1(v):
    v.scan("""\
def func():
    pass

a = func
""")
    check(v.defined_funcs, ['func'])
    check(v.unused_funcs, [])


def test_function_object2(v):
    v.scan("""\
def func():
    pass

func
""")
    check(v.defined_funcs, ['func'])
    check(v.unused_funcs, [])


def test_function1(v):
    v.scan("""\
def func1(a):
    pass

def func2(b):
    func1(b)
""")
    check(v.defined_funcs, ['func1', 'func2'])
    check(v.unused_funcs, ['func2'])


def test_function2(v):
    v.scan("""\
def func(a):
    pass

func(5)
""")
    check(v.unused_funcs, [])
    check(v.defined_funcs, ['func'])


def test_function3(v):
    v.scan("""\
def foo(a):
    pass

b = foo(5)
""")
    check(v.unused_funcs, [])
    check(v.defined_funcs, ['foo'])


def test_function_and_method1(v):
    v.scan("""\
class Bar(object):
    def func(self):
        pass

def func():
    pass

func()
""")
    check(v.defined_classes, ['Bar'])
    check(v.defined_funcs, ['func', 'func'])
    check(v.unused_classes, ['Bar'])
    check(v.unused_funcs, [])


def test_attribute1(v):
    v.scan("""\
foo.bar = 1
foo.bar = 2
""")
    check(v.unused_funcs, [])
    check(v.defined_funcs, [])
    check(v.defined_attrs, ['bar', 'bar'])
    check(v.used_attrs, [])
    check(v.unused_attrs, ['bar', 'bar'])


def test_ignored_attributes(v):
    v.scan("""\
A._ = 0
A._a = 1
A.__b = 2
A.__c__ = 3
A._d_ = 4
""")
    check(v.defined_attrs, ['_', '_a', '__b', '__c__', '_d_'])
    check(v.used_attrs, [])
    check(v.unused_attrs, ['_', '__b', '__c__', '_a', '_d_'])
    check(v.unused_vars, [])


def test_callback1(v):
    v.scan("""\
class Bar(object):
    def foo(self):
        pass

b = Bar()
b.foo
""")
    check(v.used_attrs, ['foo'])
    check(v.defined_classes, ['Bar'])
    check(v.defined_funcs, ['foo'])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])


def test_class1(v):
    v.scan("""\
class Bar(object):
    pass
""")
    check(v.used_attrs, [])
    check(v.defined_classes, ['Bar'])
    check(v.unused_classes, ['Bar'])


def test_class2(v):
    v.scan("""\
class Bar():
    pass
class Foo(Bar):
    pass
Foo()
""")
    check(v.used_attrs, [])
    check(v.used_names, ['Bar', 'Foo'])
    check(v.defined_classes, ['Bar', 'Foo'])
    check(v.unused_classes, [])


def test_class3(v):
    v.scan("""\
class Bar():
    pass
[Bar]
""")
    check(v.used_attrs, [])
    check(v.used_names, ['Bar'])
    check(v.defined_classes, ['Bar'])
    check(v.unused_classes, [])


def test_class4(v):
    v.scan("""\
class Bar():
    pass
Bar()
""")
    check(v.used_attrs, [])
    check(v.defined_classes, ['Bar'])
    check(v.unused_classes, [])


def test_class5(v):
    v.scan("""\
class Bar():
    pass
b = Bar()
""")
    check(v.used_attrs, [])
    check(v.defined_classes, ['Bar'])
    check(v.unused_classes, [])
    check(v.unused_vars, ['b'])


def test_class6(v):
    v.scan("""\
class Bar():
    pass
a = []
a.insert(0, Bar())
""")
    check(v.defined_classes, ['Bar'])
    check(v.unused_classes, [])


def test_class7(v):
    v.scan("""\
class Bar(object):
    pass

class Foo(object):
    def __init__(self):
        self.b = xyz.Bar(self)
""")
    check(v.defined_classes, ['Bar', 'Foo'])
    check(v.unused_classes, ['Foo'])


def test_method1(v):
    v.scan("""\
def __init__(self):
    self.a.foo()

class Bar(object):
    def foo(self):
        pass
""")
    check(v.defined_classes, ['Bar'])
    check(v.defined_funcs, ['foo'])
    check(v.unused_classes, ['Bar'])
    check(v.unused_funcs, [])


def test_token_types(v):
    v.scan("""\
a
b = 2
c()
x.d
""")
    check(v.defined_funcs, [])
    check(v.defined_vars, ['b'])
    check(v.used_names, ['a', 'c', 'x'])
    check(v.used_attrs, ['d'])
    check(v.unused_attrs, [])
    check(v.unused_funcs, [])
    check(v.unused_props, [])
    check(v.unused_vars, ['b'])


def test_variable1(v):
    v.scan("a = 1\nb = a")
    check(v.defined_funcs, [])
    check(v.used_names, ['a'])
    check(v.defined_vars, ['a', 'b'])
    check(v.unused_vars, ['b'])


def test_variable2(v):
    v.scan("a = 1\nc = b.a")
    check(v.defined_funcs, [])
    check(v.defined_vars, ['a', 'c'])
    check(v.used_names, ['b'])
    check(v.unused_vars, ['c'])


def test_variable3(v):
    v.scan("(a, b), c = (d, e, f)")
    check(v.defined_funcs, [])
    check(v.defined_vars, ['a', 'b', 'c'])
    check(v.used_names, ['d', 'e', 'f'])
    check(v.unused_vars, ['a', 'b', 'c'])


def test_variable4(v):
    v.scan("for a, b in func(): a")
    check(v.defined_funcs, [])
    check(v.defined_vars, ['a', 'b'])
    check(v.used_names, ['a', 'func'])
    check(v.unused_vars, ['b'])


def test_variable5(v):
    v.scan("[a for a, b in func()]")
    check(v.defined_vars, ['a', 'b'])
    check(v.used_names, ['a', 'func'])
    check(v.unused_vars, ['b'])


def test_ignored_variables(v):
    v.scan("""\
_ = 0
_a = 1
__b = 2
__c__ = 3
_d_ = 4
""")
    check(v.defined_vars, ['__b'])
    check(sorted(v.used_names), [])
    check(v.unused_vars, ['__b'])


def test_prop1(v):
    v.scan("""\
class Bar(object):
    @property
    def prop(self):
        pass

c = Bar()
c.prop
""")
    check(v.defined_classes, ['Bar'])
    check(v.defined_props, ['prop'])
    check(v.unused_classes, [])
    check(v.unused_props, [])


def test_prop2(v):
    v.scan("""\
class Bar(object):
    @property
    def prop(self):
        pass

prop = 1
""")
    check(v.defined_classes, ['Bar'])
    check(v.defined_props, ['prop'])
    check(v.defined_vars, ['prop'])
    check(v.unused_classes, ['Bar'])
    check(v.unused_props, ['prop'])


def test_object_attribute(v):
    v.scan("""\
class Bar(object):
    def __init__(self):
        self.a = []
""")
    check(v.defined_attrs, ['a'])
    check(v.defined_classes, ['Bar'])
    check(v.defined_vars, [])
    check(v.used_attrs, [])
    check(v.unused_attrs, ['a'])
    check(v.unused_classes, ['Bar'])


def test_function_names_in_test_file(v):
    v.scan("""\
def test_func():
    pass

def other_func():
    pass

class TestClass:
    pass

class BasicTestCase:
    pass

class OtherClass:
    pass
""", filename='test_function_names.py')
    check(v.defined_attrs, [])
    check(v.defined_classes, ['OtherClass'])
    check(v.defined_funcs, ['other_func'])
    check(v.defined_vars, [])
    check(v.used_attrs, [])
    check(v.used_names, [])
    check(v.unused_attrs, [])
    check(v.unused_classes, ['OtherClass'])
    check(v.unused_funcs, ['other_func'])


def test_function_names_in_normal_file(v):
    v.scan("""\
def test_func():
    pass

def other_func():
    pass

class TestClass:
    pass

class BasicTestCase:
    pass

class OtherClass:
    pass
""")
    check(v.defined_attrs, [])
    check(v.defined_classes, ['BasicTestCase', 'OtherClass', 'TestClass'])
    check(v.defined_funcs, ['test_func', 'other_func'])
    check(v.defined_vars, [])
    check(v.used_attrs, [])
    check(v.used_names, [])
    check(v.unused_attrs, [])
    check(v.unused_classes, ['BasicTestCase', 'OtherClass', 'TestClass'])
    check(v.unused_funcs, ['other_func', 'test_func'])


def test_global_attribute(v):
    v.scan("""\
# Module foo:
a = 1
if a == 1:
    pass

# Module bar:
import foo
foo.a = 2
""")
    check(v.defined_attrs, ['a'])
    check(v.defined_vars, ['a'])
    check(v.used_attrs, [])
    check(v.used_names, ['a', 'foo'])
    check(v.unused_attrs, ['a'])


def test_boolean(v):
    v.scan("""\
a = True
a
""")
    check(v.defined_vars, ['a'])
    check(v.used_names, ['a'])
    check(v.unused_vars, [])


def test_builtin_types(v):
    v.scan("""\
a = b
a = 1
a = "s"
a = object
a = False
""")
    check(v.defined_vars, ['a'] * 5)
    check(v.used_names, ['b'])
    check(v.unused_vars, ['a'] * 5)


def test_unused_args(v):
    v.scan("""\
def foo(x, y):
    return x + 1
""")
    check(v.defined_vars, ['x', 'y'])
    check(v.used_names, ['x'])
    check(v.unused_vars, ['y'])


def test_unused_kwargs(v):
    v.scan("""\
def foo(x, y=3, **kwargs):
    return x + 1
""")
    check(v.defined_vars, ['kwargs', 'x', 'y'])
    check(v.used_names, ['x'])
    check(v.unused_vars, ['kwargs', 'y'])


def test_unused_kwargs_with_odd_name(v):
    v.scan("""\
def foo(**bar):
    pass
""")
    check(v.defined_vars, ['bar'])
    check(v.used_names, [])
    check(v.unused_vars, ['bar'])


def test_unused_vararg(v):
    v.scan("""\
def foo(*bar):
    pass
""")
    check(v.defined_vars, ['bar'])
    check(v.used_names, [])
    check(v.unused_vars, ['bar'])


def test_encoding1(v):
    v.scan(u"""\
# -*- coding: utf-8 -*-
pass
""")
    assert True


def test_encoding2(v):
    v.scan(u"""\
#! /usr/bin/env python
# -*- coding: utf-8 -*-
pass
""")
    assert True


def test_multiple_definition(v):
    v.scan("""\
a = 1
a = 2
""")
    check(v.defined_vars, ['a', 'a'])
    check(v.used_names, [])
    check(v.unused_vars, ['a', 'a'])



from whitelist_utils import Whitelist

whitelist_argparser = Whitelist()
whitelist_argparser.epilog

whitelist_help_formatter = Whitelist()
whitelist_help_formatter._fill_text
whitelist_help_formatter._get_help_string



from . import v
assert v  # Silence pyflakes


def check_unreachable(v, lineno, size, name):
    assert len(v.unreachable_code) == 1
    item = v.unreachable_code[0]
    assert item.lineno == lineno
    assert item.size == size
    assert item.name == name


def test_return_assignment(v):
    v.scan("""\
def foo():
    print("Hello World")
    return
    a = 1
""")
    check_unreachable(v, 4, 1, 'return')


def test_return_multiline_return_statements(v):
    v.scan("""\
def foo():
    print("Something")
    return (something,
            that,
            spans,
            over,
            multiple,
            lines)
    print("Hello World")
""")
    check_unreachable(v, 9, 1, 'return')


def test_return_multiple_return_statements(v):
    v.scan("""\
def foo():
    return something
    return None
    return (some, statement)
""")
    check_unreachable(v, 3, 2, 'return')


def test_return_pass(v):
    v.scan("""\
def foo():
    return
    pass
    return something
""")
    check_unreachable(v, 3, 2, 'return')


def test_return_multiline_return(v):
    v.scan("""
def foo():
    return \
        "Hello"
    print("Unreachable code")
""")
    check_unreachable(v, 4, 1, 'return')


def test_return_recursive_functions(v):
    v.scan("""\
def foo(a):
    if a == 1:
        return 1
    else:
        return foo(a - 1)
        print("This line is never executed")
""")
    check_unreachable(v, 6, 1, 'return')


def test_return_semicolon(v):
    v.scan("""\
def foo():
    return; a = 1
""")
    check_unreachable(v, 2, 1, 'return')


def test_return_list(v):
    v.scan("""\
def foo(a):
    return
    a[1:2]
""")
    check_unreachable(v, 3, 1, 'return')


def test_return_continue(v):
    v.scan("""\
def foo():
    if foo():
        return True
        continue
    else:
        return False
""")
    check_unreachable(v, 4, 1, 'return')


def test_raise_assignment(v):
    v.scan("""\
def foo():
    raise ValueError
    li = []
""")
    check_unreachable(v, 3, 1, 'raise')


def test_multiple_raise_statements(v):
    v.scan("""\
def foo():
    a = 1
    raise
    raise KeyError
    # a comment
    b = 2
    raise CustomDefinedError
""")
    check_unreachable(v, 4, 4, 'raise')


def test_return_with_raise(v):
    v.scan("""\
def foo():
    a = 1
    return
    raise ValueError
    return
""")
    check_unreachable(v, 4, 2, 'return')


def test_return_comment_and_code(v):
    v.scan("""\
def foo():
    return
    # This is a comment
    print("Hello World")
""")
    check_unreachable(v, 4, 1, 'return')


def test_raise_with_return(v):
    v.scan("""\
def foo():
    a = 1
    raise
    return a
""")
    check_unreachable(v, 4, 1, 'raise')


def test_raise_error_message(v):
    v.scan("""\
def foo():
    raise SomeError("There is a problem")
    print("I am unreachable")
""")
    check_unreachable(v, 3, 1, 'raise')


def test_raise_try_except(v):
    v.scan("""\
def foo():
    try:
        a = 1
        raise
    except IOError as e:
        print("We have some problem.")
        raise
        print(":-(")
""")
    check_unreachable(v, 8, 1, 'raise')


def test_raise_with_comment_and_code(v):
    v.scan("""\
def foo():
    raise
    # This is a comment
    print("Something")
    return None
""")
    check_unreachable(v, 4, 2, 'raise')


def test_continue_basic(v):
    v.scan("""\
def foo():
    if bar():
        a = 1
    else:
        continue
        a = 2
""")
    check_unreachable(v, 6, 1, 'continue')


def test_continue_one_liner(v):
    v.scan("""\
def foo():
    for i in range(1, 10):
        if i == 5: continue
        print(1 / i)
""")
    assert v.unreachable_code == []


def test_continue_nested_loops(v):
    v.scan("""\
def foo():
    a = 0
    if something():
        foo()
        if bar():
            a = 2
            continue
            # This is unreachable
            a = 1
        elif a == 1:
            pass
        else:
            a = 3
            continue
    else:
        continue
""")
    check_unreachable(v, 9, 1, 'continue')


def test_continue_with_comment_and_code(v):
    v.scan("""\
def foo():
    if bar1():
        bar2()
    else:
        a = 1
        continue
        # Just a comment
        raise ValueError
""")
    check_unreachable(v, 8, 1, 'continue')


def test_break_basic(v):
    v.scan("""\
def foo():
    for i in range(123):
        break
        # A comment
        return
        dead = 1
""")
    check_unreachable(v, 5, 2, 'break')


def test_break_one_liner(v):
    v.scan("""\
def foo():
    for i in range(10):
        if i == 3: break
        print(i)
""")
    assert v.unreachable_code == []


def test_break_with_comment_and_code(v):
    v.scan("""\
while True:
    break
    # some comment
    print("Hello")
""")
    check_unreachable(v, 4, 1, 'break')



from . import check, v
assert v  # Silence pyflakes.


def test_old_format_string(v):
    v.scan("'%(a)s, %(b)d' % locals()")
    check(v.used_names, ['a', 'b', 'locals'])


def test_new_format_string(v):
    v.scan("'{a}, {b:0d} {c:<30} {d:.2%}'.format(**locals())")
    check(v.used_names, ['a', 'b', 'c', 'd', 'locals'])


def test_new_format_string_access(v):
    v.scan("'{a.b}, {c.d.e} {f[g]} {h[i][j]}'.format(**locals())")
    check(v.used_names, ['a', 'c', 'f', 'h', 'locals'])


def test_new_format_string_attributes(v):
    v.scan("'{a.b}, {c.d.e} {f[g]} {h[i][j].k}'.format(**locals())")
    check(v.used_names, ['a', 'c', 'f', 'h', 'locals'])
    check(v.used_attrs, ['b', 'd', 'e', 'k', 'format'])


def test_new_format_string_numbers(v):
    v.scan("'{0.b}, {0.d.e} {0[1]} {0[1][1].k}'.format('foo')")
    check(v.used_names, [])
    check(v.used_attrs, ['b', 'd', 'e', 'k', 'format'])


def test_incorrect_format_string(v):
    v.scan('"{"')
    v.scan('"{!-a:}"')
    check(v.used_names, [])
    check(v.used_attrs, [])



import ast


def _get_last_child_with_lineno(node):
    """
    Return the last direct child of `node` that has a lineno attribute,
    or None if `node` has no such children.

    Almost all node._field lists are sorted by the order in which they
    appear in source code. For some nodes however, we have to skip some
    fields that either don't have line numbers (e.g., "ctx" and "names")
    or that are in the wrong position (e.g., "decorator_list" and
    "returns"). Then we choose the first field (i.e., the field with the
    highest line number) that actually contains a node. If it contains a
    list of nodes, we return the last one.

    """
    ignored_fields = set(['ctx', 'decorator_list', 'names', 'returns'])
    fields = node._fields
    # The fields of ast.Call are in the wrong order.
    if isinstance(node, ast.Call):
        fields = ('func', 'args', 'starargs', 'keywords', 'kwargs')
    for name in reversed(fields):
        if name in ignored_fields:
            continue

        try:
            last_field = getattr(node, name)
        except AttributeError:
            continue

        # Ignore non-AST objects like "is_async", "level" and "nl".
        if isinstance(last_field, ast.AST):
            return last_field
        elif isinstance(last_field, list) and last_field:
            return last_field[-1]
    return None


def _get_last_line_number_slow(node):
    """
    A simpler and possibly more accurate (but also slower) version of
    get_last_line_number().

    It traverses all descendants of node and records the highest line
    number seen.

    This code is 1.6 times slower than count_lines() on the Python
    subset of tensorflow. It reports the same sizes for all test cases
    and the functions and classes in tensorflow. We can think about
    using the simpler version once we only compute line numbers for
    unused code.

    """
    return max(getattr(node, 'lineno', -1) for node in ast.walk(node))


def get_last_line_number(node):
    """Estimate last line number of the given AST node.

    When traversing the tree, we may see a mix of nodes with line
    numbers and nodes without line numbers. We therefore, store the
    maximum line number seen so far and report it at the end. A more
    accurate (but also slower to compute) estimate would traverse all
    children, instead of just the last one, since choosing the last one
    may lead to a path that ends with a node without line number.

    """
    max_lineno = node.lineno
    while True:
        last_child = _get_last_child_with_lineno(node)
        if last_child is None:
            return max_lineno
        else:
            try:
                max_lineno = max(max_lineno, last_child.lineno)
            except AttributeError:
                pass
        node = last_child


def count_lines(node, verbose=False):
    """
    Estimate the number of lines of the given AST node.

    The estimate is based on the line number of the last descendant of
    `node` that has a lineno attribute. Therefore, it underestimates the
    size of code ending with, e.g., multiline strings and comments.

    """
    last_lineno = get_last_line_number(node)
    if verbose:
        assert _get_last_line_number_slow(node) == last_lineno
    return last_lineno - node.lineno + 1



from . import check, v
assert v  # Silence pyflakes.


def test_import_star(v):
    v.scan("""\
from a import *
from a.b import *
""")
    check(v.defined_imports, [])
    check(v.unused_imports, [])


def test_import_from_future(v):
    v.scan("""from __future__ import division""")
    check(v.defined_imports, [])
    check(v.unused_imports, [])


def test_double_import(v):
    v.scan("""\
import foo as bar
import foo
""")
    check(v.defined_imports, ['bar', 'foo'])
    # Once the bar import is removed, the foo import will be detected.
    check(v.unused_imports, ['bar'])


def test_attribute_access(v):
    v.scan("""\
# foo.py
class Foo:
    pass

# bar.py
from foo import Foo

# main.py
import bar
bar.Foo
""")
    check(v.defined_imports, ['Foo', 'bar'])
    check(v.unused_imports, [])


def test_nested_import(v):
    v.scan("""\
import os.path
os.path.expanduser("~")
""")
    check(v.defined_imports, ['os'])
    check(v.used_names, ['os'])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


definitions = """\
class A(object):
    pass
class B(object):
    pass
def C():
    pass
D = 42
"""

imports = """\
from any_module import A
import B, C
import D
"""

aliased_imports = """\
from any_module import A as AliasA
import B as AliasB, C as AliasC
import D as AliasD
"""

uses = """\
A()
B()
C()
D()
"""

aliased_uses = """\
AliasA()
AliasB()
AliasC()
AliasD()
"""


def test_definitions(v):
    v.scan(definitions)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, [])
    check(v.defined_vars, ['D'])
    check(v.used_names, [])
    check(v.unused_classes, ['A', 'B'])
    check(v.unused_funcs, ['C'])
    check(v.unused_imports, [])
    check(v.unused_vars, ['D'])


def test_use_original(v):
    v.scan(definitions + uses)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, [])
    check(v.defined_vars, ['D'])
    check(v.used_names, ['A', 'B', 'C', 'D'])
    check(v.unused_funcs, [])
    check(v.unused_classes, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


def test_import_original(v):
    v.scan(definitions + imports)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['A', 'B', 'C', 'D'])
    check(v.defined_vars, ['D'])
    check(v.used_names, [])
    check(v.unused_classes, ['A', 'B'])
    check(v.unused_funcs, ['C'])
    check(v.unused_imports, ['A', 'B', 'C', 'D'])
    check(v.unused_vars, ['D'])


def test_import_original_use_original(v):
    v.scan(definitions + imports + uses)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['A', 'B', 'C', 'D'])
    check(v.defined_vars, ['D'])
    check(v.used_names, ['A', 'B', 'C', 'D'])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])


def test_import_original_use_alias(v):
    v.scan(definitions + imports + aliased_uses)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['A', 'B', 'C', 'D'])
    check(v.defined_vars, ['D'])
    check(v.used_names, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.unused_classes, ['A', 'B'])
    check(v.unused_funcs, ['C'])
    check(v.unused_imports, ['A', 'B', 'C', 'D'])
    check(v.unused_vars, ['D'])


def test_import_alias(v):
    v.scan(definitions + aliased_imports)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.defined_vars, ['D'])
    check(v.used_names, ['A', 'B', 'C', 'D'])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.unused_vars, [])


def test_import_alias_use_original(v):
    v.scan(definitions + aliased_imports + uses)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.defined_vars, ['D'])
    check(v.used_names, ['A', 'B', 'C', 'D'])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.unused_vars, [])


def test_import_alias_use_alias(v):
    v.scan(definitions + aliased_imports + aliased_uses)
    check(v.defined_classes, ['A', 'B'])
    check(v.defined_funcs, ['C'])
    check(v.defined_imports, ['AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.defined_vars, ['D'])
    check(
        v.used_names,
        ['A', 'B', 'C', 'D', 'AliasA', 'AliasB', 'AliasC', 'AliasD'])
    check(v.unused_classes, [])
    check(v.unused_funcs, [])
    check(v.unused_imports, [])
    check(v.unused_vars, [])



import ast
import sys

import pytest
from vulture import utils

from . import check_unreachable
from . import v
assert v  # Silence pyflakes


def check_condition(code, result):
    condition = ast.parse(code, mode='eval')
    assert utils._evaluate_condition(condition) == result


def test_false():
    check_condition('False', False)
    check_condition('None', False)
    check_condition("0", False)
    if sys.version_info > (3, 0):
        check_condition("1 - 1", False)


def test_empty():
    check_condition("''", False)
    check_condition("[]", False)
    check_condition("{}", False)


def test_true():
    check_condition("True", True)
    check_condition("2", True)
    check_condition("['foo', 'bar']", True)
    check_condition("{'a': 1, 'b': 2}", True)


def test_errors():
    conditions = [
        'foo',
        'foo and False',
        '__name__ == "__main__"',
        'False and 1',
        'False and foo',
        'chr(-1)',
        'getattr(True, "foo")',
        'hasattr(str, "foo")',
        'isinstance(True, True)',
        'not False']
    for condition in conditions:
        with pytest.raises(ValueError) as e:
            check_condition(condition, False)
        assert e.match("Condition cannot be evaluated")


def test_while(v):
    v.scan("""\
while False:
    pass
""")
    check_unreachable(v, 1, 2, 'while')


def test_while_nested(v):
    v.scan("""\
while True:
    while False:
        pass
""")
    check_unreachable(v, 2, 2, 'while')


def test_if_false(v):
    v.scan("""\
if False:
    pass
""")
    check_unreachable(v, 1, 2, 'if')


def test_elif_false(v):
    v.scan("""\
if bar():
    pass
elif False:
    print("Unreachable")
""")
    check_unreachable(v, 3, 2, 'if')


def test_nested_if_statements_false(v):
    v.scan("""\
if foo():
    if bar():
        pass
    elif False:
        print("Unreachable")
        pass
    elif something():
        print("Reachable")
    else:
        pass
else:
    pass
""")
    check_unreachable(v, 4, 3, 'if')


def test_if_false_same_line(v):
    v.scan("""\
if False: a = 1
else: c = 3
""")
    check_unreachable(v, 1, 1, 'if')


def test_if_true(v):
    v.scan("""\
if True:
    a = 1
    b = 2
else:
    c = 3
    d = 3
""")
    # For simplicity, we don't report the "else" line as dead code.
    check_unreachable(v, 5, 2, 'else')


def test_if_true_same_line(v):
    v.scan("""\
if True:
    a = 1
    b = 2
else: c = 3
d = 3
""")
    check_unreachable(v, 4, 1, 'else')


def test_nested_if_statements_true(v):
    v.scan("""\
if foo():
    if bar():
        pass
    elif True:
        if something():
            pass
        else:
            pass
    elif something_else():
        print("foo")
    else:
        print("bar")
else:
    pass
""")
    check_unreachable(v, 9, 4, 'else')



import glob
import os.path
import subprocess
import sys

DIR = os.path.dirname(os.path.abspath(__file__))
REPO = os.path.dirname(DIR)
WHITELISTS = glob.glob(os.path.join(REPO, 'vulture', 'whitelists', '*.py'))


def call_vulture(args, **kwargs):
    core = 'vulture'
    if sys.version_info < (2, 7):
        core = 'vulture.core'
    return subprocess.call(
        [sys.executable, '-m', core] + args, cwd=REPO, **kwargs)


def test_module_with_explicit_whitelists():
    assert call_vulture(['vulture/'] + WHITELISTS) == 0


def test_module_with_implicit_whitelists():
    assert call_vulture(['vulture/']) == 0


def test_module_without_whitelists():
    assert call_vulture(['vulture/', '--exclude', 'whitelists']) == 1


def test_missing_file():
    assert call_vulture(['missing.py']) == 1


def test_tests():
    assert call_vulture(['tests/']) == 0


def test_whitelists_with_python():
    for whitelist in WHITELISTS:
        assert subprocess.call([sys.executable, whitelist], cwd=REPO) == 0


def test_pyc():
    assert call_vulture(['missing.pyc']) == 1


def test_sort_by_size():
    assert call_vulture(['vulture/', '--sort-by-size']) == 0


def test_min_confidence():
    assert call_vulture([
        'vulture/core.py', '--exclude', 'whitelists',
        '--min-confidence', '100']) == 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# vulture - Find dead code.
#
# Copyright (c) 2012-2017 Jendrik Seipp (jendrikseipp@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function

import ast
from fnmatch import fnmatchcase
import optparse
import os
import pkgutil
import re
import string
import sys

from vulture import lines
from vulture import utils

__version__ = '0.24'

DEFAULT_CONFIDENCE = 60

# The ast module in Python 2 trips over "coding" cookies, so strip them.
ENCODING_REGEX = re.compile(
    r"^[ \t\v]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+).*?$", flags=re.M)

IGNORED_VARIABLE_NAMES = set(['object', 'self'])
# True and False are NameConstants since Python 3.4.
if sys.version_info < (3, 4):
    IGNORED_VARIABLE_NAMES |= set(['True', 'False'])


def _get_unused_items(defined_items, used_names):
    unused_items = [item for item in set(defined_items)
                    if item.name not in used_names]
    unused_items.sort(key=lambda item: item.name.lower())
    return unused_items


def _is_special_name(name):
    return name.startswith('__') and name.endswith('__')


def _is_test_file(filename):
    name = os.path.basename(filename)
    return any(
        fnmatchcase(name, pattern)
        for pattern in ['test*.py', '*_test.py'])


def _ignore_class(filename, class_name):
        return _is_test_file(filename) and 'Test' in class_name


def _ignore_import(_filename, import_name):
    # Ignore star-imported names, since we can't detect whether they are used.
    return import_name == '*'


def _ignore_function(filename, function_name):
    return (
        _is_special_name(function_name) or
        (function_name.startswith('test_') and _is_test_file(filename)))


def _ignore_variable(filename, varname):
    """
    Ignore _ (Python idiom), _x (pylint convention) and
    __x__ (special variable or method), but not __x.
    """
    return (
        varname in IGNORED_VARIABLE_NAMES or
        (varname.startswith('_') and not varname.startswith('__')) or
        _is_special_name(varname))


class Item(object):
    """
    Hold the name, type and location of defined code.
    """
    def __init__(self, name, typ, filename, lineno, size=1, message='',
                 confidence=DEFAULT_CONFIDENCE):
        self.name = name
        self.typ = typ
        self.filename = filename
        self.lineno = lineno
        self.size = size
        self.message = message or "unused {typ} '{name}'".format(**locals())
        self.confidence = confidence

    def _tuple(self):
        return (self.filename, self.lineno, self.name)

    def __repr__(self):
        return repr(self.name)

    def __eq__(self, other):
        return self._tuple() == other._tuple()

    def __hash__(self):
        return hash(self._tuple())


class Vulture(ast.NodeVisitor):
    """Find dead code."""
    def __init__(self, exclude=None, verbose=False, sort_by_size=False,
                 min_confidence=0):
        if not 0 <= min_confidence <= 100:
            raise ValueError('min_confidence must be between 0 and 100.')
        self.exclude = []
        self.sort_by_size = sort_by_size
        self.min_confidence = min_confidence

        for pattern in exclude or []:
            if not any(char in pattern for char in ['*', '?', '[']):
                pattern = '*{0}*'.format(pattern)
            self.exclude.append(pattern)

        self.verbose = verbose

        def get_list(typ):
            return utils.LoggingList(typ, self.verbose)

        def get_set(typ):
            return utils.LoggingSet(typ, self.verbose)

        self.defined_attrs = get_list('attribute')
        self.defined_classes = get_list('class')
        self.defined_funcs = get_list('function')
        self.defined_imports = get_list('import')
        self.defined_props = get_list('property')
        self.defined_vars = get_list('variable')
        self.unreachable_code = get_list('unreachable_code')

        self.used_attrs = get_set('attribute')
        self.used_names = get_set('name')

        self.filename = ''
        self.code = []
        self.found_dead_code_or_error = False

    def scan(self, code, filename=''):
        code = ENCODING_REGEX.sub("", code, count=1)
        self.code = code.splitlines()
        self.filename = filename
        try:
            node = ast.parse(code, filename=self.filename)
        except SyntaxError as err:
            text = ' at "{0}"'.format(err.text.strip()) if err.text else ''
            print('{0}:{1:d}: {2}{3}'.format(
                utils.format_path(filename), err.lineno, err.msg, text),
                file=sys.stderr)
            self.found_dead_code_or_error = True
        except (TypeError, ValueError) as err:
            # Python < 3.5 raises TypeError and Python >= 3.5 raises
            # ValueError if source contains null bytes.
            print('{0}: invalid source code "{1}"'.format(
                utils.format_path(filename), err), file=sys.stderr)
            self.found_dead_code_or_error = True
        else:
            self.visit(node)

    def _get_modules(self, paths, toplevel=True):
        """Take files from the command line even if they don't end with .py."""
        modules = []
        for path in paths:
            path = os.path.abspath(path)
            if toplevel and path.endswith('.pyc'):
                sys.exit('.pyc files are not supported: {0}'.format(path))
            if os.path.isfile(path) and (path.endswith('.py') or toplevel):
                modules.append(path)
            elif os.path.isdir(path):
                subpaths = [
                    os.path.join(path, filename)
                    for filename in sorted(os.listdir(path))]
                modules.extend(self._get_modules(subpaths, toplevel=False))
            elif toplevel:
                sys.exit('Error: {0} could not be found.'.format(path))
        return modules

    def scavenge(self, paths):
        def exclude(name):
            return any(fnmatchcase(name, pattern) for pattern in self.exclude)

        for module in self._get_modules(paths):
            if exclude(module):
                self._log('Excluded:', module)
                continue

            self._log('Scanning:', module)
            try:
                module_string = utils.read_file(module)
            except utils.VultureInputException as err:
                print(
                    'Error: Could not read file {module} - {err}\n'
                    'Try to change the encoding to UTF-8.'.format(**locals()),
                    file=sys.stderr)
                self.found_dead_code_or_error = True
            else:
                self.scan(module_string, filename=module)

        unique_imports = set(item.name for item in self.defined_imports)
        for import_name in unique_imports:
            path = os.path.join('whitelists', import_name) + '.py'
            if exclude(path):
                self._log('Excluded whitelist:', path)
            else:
                try:
                    module_data = pkgutil.get_data('vulture', path)
                    self._log('Included whitelist:', path)
                except IOError:
                    # Most imported modules don't have a whitelist.
                    continue
                if module_data is None:
                    sys.exit('Error: Please use "python -m vulture".')
                module_string = module_data.decode("utf-8")
                self.scan(module_string, filename=path)

    def get_unused_code(self):
        """
        Return ordered list of unused Item objects.
        """
        def by_size(item):
            return item.size

        def by_name(item):
            return (item.filename.lower(), item.lineno)

        unused_code = (self.unused_attrs + self.unused_classes +
                       self.unused_funcs + self.unused_imports +
                       self.unused_props + self.unused_vars +
                       self.unreachable_code)

        confidently_unused = [obj for obj in unused_code
                              if obj.confidence >= self.min_confidence]

        return sorted(confidently_unused,
                      key=by_size if self.sort_by_size else by_name)

    def report(self):
        """
        Print ordered list of Item objects to stdout.
        """
        for item in self.get_unused_code():
            if self.sort_by_size:
                line_format = 'line' if item.size == 1 else 'lines'
                size_report = ', {0:d} {1}'.format(item.size, line_format)
            else:
                size_report = ''

            print("{0}:{1:d}: {2} ({3}% confidence{4})".format(
                utils.format_path(item.filename), item.lineno,
                item.message, item.confidence, size_report))
            self.found_dead_code_or_error = True
        return self.found_dead_code_or_error

    @property
    def unused_classes(self):
        return _get_unused_items(
            self.defined_classes,
            self.used_attrs | self.used_names)

    @property
    def unused_funcs(self):
        return _get_unused_items(
            self.defined_funcs,
            self.used_attrs | self.used_names)

    @property
    def unused_imports(self):
        return _get_unused_items(
            self.defined_imports,
            self.used_names | self.used_attrs)

    @property
    def unused_props(self):
        return _get_unused_items(self.defined_props, self.used_attrs)

    @property
    def unused_vars(self):
        return _get_unused_items(
            self.defined_vars,
            self.used_attrs | self.used_names)

    @property
    def unused_attrs(self):
        return _get_unused_items(self.defined_attrs, self.used_attrs)

    def _log(self, *args):
        if self.verbose:
            print(*args)

    def _add_aliases(self, node):
        assert isinstance(node, (ast.Import, ast.ImportFrom))
        for name_and_alias in node.names:
            # Store only top-level module name ("os.path" -> "os").
            # We can't detect when "os.path" is used.
            name = name_and_alias.name.partition('.')[0]
            alias = name_and_alias.asname
            self._define(
                self.defined_imports, alias or name, node.lineno,
                confidence=90, ignore=_ignore_import)
            if alias is not None:
                self.used_names.add(name_and_alias.name)

    def _define(self, collection, name, lineno, size=1, message='',
                confidence=DEFAULT_CONFIDENCE, ignore=None):
        typ = collection.typ
        if ignore and ignore(self.filename, name):
            self._log('Ignoring {typ} "{name}"'.format(**locals()))
        else:
            collection.append(
                Item(name, typ, self.filename, lineno, size=size,
                     message=message, confidence=confidence))

    def _define_variable(self, name, lineno, confidence=DEFAULT_CONFIDENCE):
        self._define(self.defined_vars, name, lineno, confidence=confidence,
                     ignore=_ignore_variable)

    def visit_alias(self, node):
        """
        Use the methods below for imports to have access to line numbers
        and to filter imports from __future__.
        """
        pass

    def visit_arg(self, node):
        """Function argument. Python 3 only. Has lineno since Python 3.4"""
        self._define_variable(node.arg, getattr(node, 'lineno', -1),
                              confidence=100)

    def visit_Attribute(self, node):
        if isinstance(node.ctx, ast.Store):
            size = lines.count_lines(node) if self.sort_by_size else 1
            self._define(self.defined_attrs, node.attr, node.lineno, size=size)
        elif isinstance(node.ctx, ast.Load):
            self.used_attrs.add(node.attr)

    def visit_ClassDef(self, node):
        size = lines.count_lines(node) if self.sort_by_size else 1
        self._define(
            self.defined_classes, node.name, node.lineno,
            size=size, ignore=_ignore_class)

    def visit_FunctionDef(self, node):
        size = lines.count_lines(node) if self.sort_by_size else 1
        for decorator in node.decorator_list:
            if getattr(decorator, 'id', None) == 'property':
                self._define(
                    self.defined_props, node.name, node.lineno, size=size)
                break
        else:
            # Function is not a property.
            self._define(
                self.defined_funcs, node.name, node.lineno, size=size,
                ignore=_ignore_function)

        # Detect *args and **kwargs parameters. Python 3 recognizes them
        # in visit_Name. For Python 2 we use this workaround. We can't
        # use visit_arguments, because its node has no lineno.
        for param in [node.args.vararg, node.args.kwarg]:
            if param and isinstance(param, str):
                self._define_variable(param, node.lineno, confidence=100)

    def visit_If(self, node):
        else_body = getattr(node, 'orelse')
        if utils.condition_is_always_true(node.test) and else_body:
            if self.sort_by_size:
                size = (lines.get_last_line_number(else_body[-1]) -
                        else_body[0].lineno + 1)
            else:
                size = 1
            self._define(self.unreachable_code, 'else', else_body[0].lineno,
                         size=size,
                         message="unreachable 'else' block",
                         confidence=100)
        elif utils.condition_is_always_false(node.test):
            if self.sort_by_size:
                size = (lines.get_last_line_number(node.body[-1]) -
                        node.lineno + 1)
            else:
                size = 1
            self._define(self.unreachable_code, 'if', node.lineno,
                         size=size,
                         message="unsatisfiable 'if' condition",
                         confidence=100)

    def visit_Import(self, node):
        self._add_aliases(node)

    def visit_ImportFrom(self, node):
        if node.module != '__future__':
            self._add_aliases(node)

    def visit_Name(self, node):
        if (isinstance(node.ctx, ast.Load) and
                node.id not in IGNORED_VARIABLE_NAMES):
            self.used_names.add(node.id)
        elif isinstance(node.ctx, (ast.Param, ast.Store)):
            self._define_variable(node.id, node.lineno)

    def visit_Str(self, node):
        """
        Parse variable names in format strings:

        '%(my_var)s' % locals()
        '{my_var}'.format(**locals())

        """
        # Old format strings.
        self.used_names |= set(re.findall(r'\%\((\w+)\)', node.s))

        def is_identifier(s):
            return bool(re.match(r'[a-zA-Z_][a-zA-Z0-9_]*', s))

        # New format strings.
        parser = string.Formatter()
        try:
            names = [name for _, name, _, _ in parser.parse(node.s) if name]
        except ValueError:
            # Invalid format string.
            names = []

        for field_name in names:
            # Remove brackets and contents: "a[0][b].c[d].e" -> "a.c.e".
            # "a.b.c" -> name = "a", attributes = ["b", "c"]
            name_and_attrs = re.sub(r'\[\w*\]', '', field_name).split('.')
            name = name_and_attrs[0]
            if is_identifier(name):
                self.used_names.add(name)
            for attr in name_and_attrs[1:]:
                if is_identifier(attr):
                    self.used_attrs.add(attr)

    def visit_While(self, node):
        if utils.condition_is_always_false(node.test):
            self._define(self.unreachable_code, 'while', node.lineno,
                         size=lines.count_lines(node) if self.sort_by_size
                         else 1,
                         message="unsatisfiable 'while' condition",
                         confidence=100)

    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if self.verbose:
            lineno = getattr(node, 'lineno', 1)
            line = self.code[lineno - 1] if self.code else ''
            self._log(lineno, ast.dump(node), line)
        if visitor:
            visitor(node)
        return self.generic_visit(node)

    def _handle_ast_list(self, ast_list):
        """
        Find unreachable nodes in the given sequence of ast nodes.
        """
        for index, node in enumerate(ast_list):
            if isinstance(node, (ast.Break, ast.Continue, ast.Raise,
                                 ast.Return)):
                try:
                    first_unreachable_node = ast_list[index + 1]
                except IndexError:
                    continue
                class_name = node.__class__.__name__.lower()
                self._define(
                    self.unreachable_code,
                    class_name,
                    first_unreachable_node.lineno,
                    size=lines.get_last_line_number(ast_list[-1]) -
                    first_unreachable_node.lineno + 1,
                    message="unreachable code after '{class_name}'".format(
                        **locals()),
                    confidence=100)
                return

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for _, value in ast.iter_fields(node):
            if isinstance(value, list):
                self._handle_ast_list(value)
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)


def _parse_args():
    def csv(option, _, value, parser):
        setattr(parser.values, option.dest, value.split(','))
    usage = """\
usage: %prog [options] PATH [PATH ...]

Paths may be Python files or directories. For each directory vulture
analyzes all contained *.py files.
"""
    version = "vulture {0}".format(__version__)
    parser = optparse.OptionParser(usage=usage, version=version)
    parser.add_option(
        '--exclude', action='callback', callback=csv, metavar='PATTERN',
        type='string', default=[],
        help=(
            'Comma-separated list of paths to ignore (e.g.,'
            ' *settings.py,docs/*.py). PATTERNs can contain globbing'
            ' characters (*, ?, [, ]). Treat PATTERNs without globbing'
            ' characters as *PATTERN*.'))
    parser.add_option(
        "--sort-by-size", action="store_true",
        help="Sort unused functions and classes by their lines of code")
    parser.add_option('-v', '--verbose', action='store_true')
    parser.add_option(
        '--min-confidence', action='store', type='int', default=0, help=(
            'Minimum confidence (between 0 and 100) for code to be'
            ' reported as unused.'))
    options, args = parser.parse_args()
    return options, args


def main():
    options, args = _parse_args()
    vulture = Vulture(exclude=options.exclude, verbose=options.verbose,
                      sort_by_size=options.sort_by_size,
                      min_confidence=options.min_confidence)
    vulture.scavenge(args)
    sys.exit(vulture.report())


# Only useful for Python 2.6 which doesn't support "python -m vulture".
if __name__ == '__main__':
    main()



from whitelist_utils import Whitelist

whitelist_test_case = Whitelist()
whitelist_test_case.setUp
whitelist_test_case.tearDown
whitelist_test_case.setUpClass
whitelist_test_case.tearDownClass
whitelist_test_case.run
whitelist_test_case.skipTest
whitelist_test_case.subTest
whitelist_test_case.debug



#! /usr/bin/env python

import sys

import setuptools
from setuptools.command.test import test as TestCommand

from vulture import __version__


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)

        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        sys.exit(pytest.main(self.test_args))


setuptools.setup(
    name='vulture',
    version=__version__,
    description='Find dead code',
    long_description='\n\n'.join(
        [open('README.rst').read(), open('NEWS.rst').read()]),
    keywords='dead-code-removal',
    author='Jendrik Seipp',
    author_email='jendrikseipp@gmail.com',
    url='https://github.com/jendrikseipp/vulture',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Quality Assurance'
    ],
    entry_points={
        'console_scripts': ['vulture = vulture.core:main'],
    },
    tests_require=['pytest', 'pytest-cov'],
    cmdclass={'test': PyTest},
    packages=setuptools.find_packages(exclude=['tests']),
    package_data={'vulture': ['whitelists/*.py']},
)
